/** smtpd.c
*
*/
char version_txt[] =
    "$VER: SMTPd 1.9 (18.4.2000)";
/*         =====================
*
*        SMTP mail server.
*/
static char *usage =
   "Usage: smtpd [-h] [-r] [-c] [-l] [-m <mail-dir>] [-u] [-d <info-level>] [-s <sound-file>]\n"
   "where:\n"
   "-h   gives this message\n"
   "-r   turns on mail relaying\n"
   "-c   turns off checking whether connecting host IP resolves to hostname given\n"
   "-l   log debug info to a file rather than stdout (use with -d)\n"
   "-m   sets the incoming mail directory (default: /boot/home/mail/in)\n"
   "-u   turns off checking if the user exists on the system\n"
   "-d   turns on debugging information, with info levels 0 through 7\n"
   "-s   sets a soundfile to use when mail arrives\n";
/*
*
*   By W.G.J Langeveld.
*   Fixes and multi-threadedness in version 1.8 by David Reid.
*   Allowing localhost (with some security), relaying and logging
*   in version 1.9 kindly provided by David Reid.
*
**/
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <sys/dir.h>
#include <pwd.h>
#include <kernel/OS.h>

#include <sys/socket.h>

#include <TypeConstants.h>

#include <fs_attr.h>
#include <parsedate.h>

#include "mailbeep.h"
#include "setmimetype.h"
#include "readn.h"
#include "mxget.h"
/*
*   Prototypes.
*/
void dprintf(long, char *, ...);
long putline(long, char *, ...);
long getline(long, char *);
long extractpath(char *, char *);
char* collectfile(long, char *, char *, char *, char *, char *, char *);
void SetParsedAttrs(FILE *fp);
char *findopt(int argc, char **argv, char opt);
void fwritebr(const char *, FILE *, int);
int32 process_connection(void*);
/*
*   Info levels:
*/
#define SMERR_FATAL    0
#define SMERR_SEVERE   1
#define SMERR_ERROR    2
#define SMERR_WARNING  3
#define SMERR_INFO     4
#define SMERR_DETAIL   6

static long infolevel = SMERR_SEVERE;
FILE *smtplog = NULL;
/*
*   Some defines
*/
#define MAXHOSTNAMELEN   64
#define MAXLINE        4096

#define SMTP_LOG_FILE     "/var/log/smtp.log" 
/*
*   Multi-thread support
*/
typedef struct {
   int sock;
   struct sockaddr_in sa;
} thread_data;

/*
*   Commands known by this server
*/
#define MKL(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))

#define HELO MKL('H','E','L','O')
#define MAIL MKL('M','A','I','L')
#define RCPT MKL('R','C','P','T')
#define DATA MKL('D','A','T','A')
#define RSET MKL('R','S','E','T')
#define SEND MKL('S','E','N','D')
#define SOML MKL('S','O','M','L')
#define SAML MKL('S','A','M','L')
#define VRFY MKL('V','R','F','Y')
#define EXPN MKL('E','X','P','N')
#define HELP MKL('H','E','L','P')
#define NOOP MKL('N','O','O','P')
#define QUIT MKL('Q','U','I','T')
#define TURN MKL('T','U','R','N')
/*
*   DBUG is not part of the spec. It will generate a "not implemented" error,
*   but will in fact set the "infolevel" to a high value.
*/
#define DBUG MKL('D','B','U','G')

static char myhostname[MAXHOSTNAMELEN + 4];
static char *maildir     = NULL;
static char *help        = NULL;
static char *nousercheck = NULL;
static char *debug       = NULL;
static char *soundfile   = NULL;
static char *relaying    = NULL;
static char *nochecking  = NULL;
static char *logging     = NULL;

/**
*
*   Main program. Executed from inetd.
*
**/
int main(int argc, char **argv)
{
   long  ss = -1, ll, s;
   int port, n;
   struct sockaddr_in sin, sin2;
   int fp;
   
   char bf[MAXLINE + 16], *buff, *ptr1;

   struct hostent *he;
   struct passwd  *pw;
/*
*   Arguments
*/
   help         = findopt(argc, argv, 'h');
   if (help) {
      printf(usage);
      exit(0);
   }

   maildir      = findopt(argc, argv, 'm');
   if (maildir == NULL) maildir = "/boot/home/mail/in";

   nousercheck  = findopt(argc, argv, 'u');
   debug        = findopt(argc, argv, 'd');
   soundfile    = findopt(argc, argv, 's');
   relaying     = findopt(argc, argv, 'r');
   nochecking   = findopt(argc, argv, 'c');
   logging      = findopt(argc, argv, 'l');
/*
*   Get our own host name and user name.
*/
   gethostname(myhostname, MAXHOSTNAMELEN);
   if (myhostname[0] == '\0') {
      dprintf(SMERR_FATAL, "Smtpd: can't find local host name\n");
      exit(0);
   }

   he = gethostbyname(myhostname);
   strcpy(myhostname, he->h_name);

   if (myhostname[0] == '\0') {
      dprintf(SMERR_FATAL, "Smtpd: invalid local host name\n");
      exit(0);
   }

   if (logging && ((smtplog = fopen(SMTP_LOG_FILE, "a+")) == NULL)) {
      dprintf(SMERR_FATAL,"Couldn't open the log file!");
   }
/*
*   Check for info level flag
*/
   if (debug) infolevel = atol(debug);

   if (infolevel < 0) infolevel = 0;
   if (infolevel > 7) infolevel = 7;
/*
*   Set up TCP server
*/
   ss = socket(AF_INET, SOCK_STREAM, 0);
   if (ss < 0) {
      dprintf(SMERR_FATAL, "Unable to obtain a socket\n");
      exit(20);
   }
/*
*   We're an SMTP server.
*/ 
   port = 25;

   sin.sin_family      = AF_INET;
   sin.sin_addr.s_addr = INADDR_ANY;
   sin.sin_port        = htons(port);

   if (bind(ss, (struct sockaddr *) &sin, sizeof (sin)) < 0) {
      dprintf(SMERR_FATAL, "%ld> Unable to bind\n", s);
      exit(20);
   }

   if (listen(ss, 5) < 0) {
      dprintf(SMERR_FATAL, "%ld> Listen failure\n", s);
      exit(20);
   }

   for (;;) {
      thread_id t;
      thread_data *td = malloc(sizeof(thread_data));

      n = sizeof(sin2);

      s = accept(ss, (struct sockaddr *) &sin2, &n);
      if (s < 0) {
         dprintf(SMERR_FATAL, "%ld> Accept failure\n", ss);
         exit(20);
      }
      td->sock = s;
      td->sa = sin2;
      t = spawn_thread(process_connection, "smtp_worker", B_NORMAL_PRIORITY, (void *) td);
      if (t > 0) resume_thread(t); 
   }

   return;
}


/**
*
*   Find an option in the option array.
*
**/
char *findopt(int argc, char **argv, char opt)
{
   int i;
   for (i = 1; i < argc; i++) {
      if (argv[i][0] == '-') {
         if (argv[i][1] == opt) {
            if (argv[i][2] != '\0') {
               return(&argv[i][2]);
            }
            else {
               i++;
               if (argc > i) return(argv[i]);
               else          return("");
            }
         }
      }
   }
   return(NULL);
}


/**
*
*   Get a line from remote
*
**/
long getline(long s, char *buf)
{
   long n;

   n = readline(s, buf, MAXLINE);
   if (n >= 0) {
      dprintf(SMERR_INFO,  "Smtpd %ld- R: %s", s, buf);
   }
   else {
      dprintf(SMERR_WARNING, "Smtpd %ld: remote terminated early\n", s);
   }

   return(n);
}

/**
*
*   Send a line to the remote
*
**/
long putline(long s, char *fmt, ...)
{
   va_list args;
   int len;
   char buf[260];

   if (fmt) {
      va_start(args, fmt);
      vsprintf(buf, fmt, args);

      dprintf(SMERR_INFO, "Smtpd %ld- S: %s\n", s, buf);

      len = strlen(buf);
      buf[len++] = '\r';
      buf[len++] = '\n';

      writen(s, buf, len);

      va_end(args);
   }
}

/**
*
*   Extract a path specification
*
**/
long extractpath(char *s, char *t)
{
   char *ptr1, *ptr2;
   long len, retval = 0;

   *s = '\0';

   ptr1 = strchr(t, '<');
   ptr2 = strchr(t, '>');

   if ((ptr1 == NULL) || (ptr2 == NULL)) {
      strcpy(t, "501 Illegal path");
   }
   else {
      ptr1++;
      *ptr2 = '\0';
      len = strlen(ptr1);
      if (len > 256) {
         strcpy(t, "500 Line too long");
      }
      else {
         strcpy(s, ptr1);
         strcpy(t, "250 OK");
         retval = 1;
      }
   }

   return(retval);
}

/**
*
*   Collect the mail message
*
**/
char * collectfile(long s, char *maildir, char *myhostname,
                         char *rpath, char *hostname, char *recipient, char *buff)
{
   register int len, retval = 1;
   char *line;
   FILE *fp;
   char datebuf[100];
   char datebuf2[100];
   struct Window *oldw;
   unsigned long number;
   unsigned long seqno = 0;
   char filename[1024];
   struct tm *tmptr;
   int off;
   char *offchar;
/*
*   Create the file with a unique name.
*/
   len = strlen(maildir);
   number = time(NULL);

   while (1) {
      sprintf(filename, "%s%s%s_%ld_%ld", maildir,
              ((maildir[len - 1] != '/') && (maildir[len - 1] != ':')) ? "/" : "",
              rpath, number, seqno);

      dprintf(SMERR_DETAIL, "Smtpd %ld: opening file %s\n", s, filename);
      if (access(filename, F_OK) != 0) break;

      seqno++;
   }

   fp = fopen(filename, "w+");
   if (fp == NULL) return(0);

   sprintf(buff, "Return-Path: <%s>", rpath);
   fwritebr(buff, fp, 72);
   dprintf(SMERR_DETAIL, buff);

   tmptr = localtime(&number);
   strftime(datebuf, sizeof(datebuf), "%a, %d %b %Y %H:%M:%S %%s%%02d00 (%Z)", tmptr);

   off = tmptr->tm_gmtoff/3600;
   if (off < 0) {
      offchar = "-";
      off = -off;
   }
   else {
      offchar = "+";
   }
 
   sprintf(datebuf2, datebuf, offchar, off);
   
   sprintf(buff, "Received: from %s by %s with BeOS SMTPd#%05ld for %s on %s",
                      hostname, myhostname, number, recipient, datebuf2);
   

   fwritebr(buff, fp, 72);
   dprintf(SMERR_DETAIL, buff);

   while (1) {
/*
*   Our "readline" turns cr/lf into lf.
*/
      readline(s, buff, MAXLINE);
      if (buff[0] == '.') {
         if (buff[1] == '\n' || buff[1] == '\r') break;
         line = &buff[1];
      }
      else {
         line = buff;
      }

      len = strlen(line);
/*
*   Some mail readers on BeOS need the cr/lf pair, so put it back in
*/
      line[len - 1] = '\r';
      line[len++]   = '\n';
      if (fwrite(line, 1, len, fp) != len) retval = 0;
   }

   if (retval == 0) dprintf(SMERR_WARNING, "Smtpd %ld: write error", s);

   fclose(fp);

   return(retval == 0 ? NULL : filename);
}

void do_mail_local(char * filename)
{
   FILE *fp;
   
   set_mime_type(filename, "text/x-email");

   fp = fopen(filename, "r+");
   SetParsedAttrs(fp);
   fclose (fp);
}

struct hostent* get_mail_host(char *domain)
{
   struct mxentry *mxr, *mxp;
   char *mxname = NULL;
   struct hostent *he;
   int lpref;

   if (relaying == NULL) return(NULL);

   mxr = getmxrecords(domain);
   if (mxr == NULL) return(NULL);

/*
*   loop through all the mx records we got and look for a matching
*   domain name...
*/
   for (mxp = mxr; mxp->name; mxp++) {
      if (strcasecmp(domain, mxp->name) == 0) mxname = strdup(mxp->name);
   }
  
/*
*   OK, so we still haven't got a match, let's go for the
*   one with the lowest preference...
*/
   lpref = 100;
   if (mxname == NULL) {
      for (mxp = mxr; mxp->name; mxp++){
         if (mxp->pref < lpref) {
            mxname = strdup(mxp->name);
            lpref = mxp->pref;
         }
      }
   }   
/*
*   did we succeed?
*/
   if (mxname == NULL) return (NULL);
     
   dprintf(SMERR_DETAIL, "using %s as mail host", mxname);

   he = gethostbyname(mxname);
   return he; 
}

void send_remote(int sock, char *buff)
{
   putline(sock, buff);  
   dprintf(SMERR_DETAIL, ">> %s\n", buff);
   return;
}

int check_remote_response(int sock, char *resp)
{
   char buff[1024];
   int bytes;
   
   if ((bytes = recv (sock, buff, 1024, 0)) <= 0) return (0);
   
   buff[bytes -1] = '\0'; /* remove carriage return */
   dprintf(SMERR_DETAIL, "<< %s\n", buff);
   
   if (strncmp(buff, resp, strlen(resp)) == 0) return(1);

   return(0);     
}   
   
int do_remote_mail(char *filename, char *rpath, char *recipient,
                   struct sockaddr_in *server)
{
   int s = socket(AF_INET, SOCK_STREAM ,0);
   int r;
   char outbuff[1024], inbuff[1024];
   int fp, bytes;
      
   if (s < 0) {
      dprintf(SMERR_FATAL, "failed to get a socket to use for remote server!");
      unlink(filename);
      return(-1);
   }

   dprintf(SMERR_DETAIL, "trying to connect now...");
   
   r = connect(s, (struct sockaddr *) server, sizeof(server));
   if (r < 0) {
      dprintf(SMERR_FATAL, "couldn't connect to remote host!");
      unlink(filename);
      return(-1);
   }

   dprintf(SMERR_DETAIL, "connected!");
   
   if (!check_remote_response(s, "220")) goto end_send_attempt;
   
   sprintf(outbuff, "helo %s\0", myhostname);
   send_remote(s, outbuff);
   if (!check_remote_response(s, "250")) goto end_send_attempt;
     
   sprintf(outbuff, "mail from: <%s>\0", rpath);
   send_remote(s, outbuff);
   if (!check_remote_response(s, "250")) goto end_send_attempt;

   sprintf(outbuff, "rcpt to: <%s>\0", recipient);
   send_remote(s, outbuff);
   if (!check_remote_response(s, "250")) goto end_send_attempt;

   sprintf(outbuff, "data");
   send_remote(s, outbuff);
   if (!check_remote_response(s, "354")) goto end_send_attempt;

   fp = open(filename, O_RDONLY);
   while ((bytes = read(fp, outbuff, 1024)) > 0) {
      send(s, outbuff, bytes, 0);
   }
   close(fp);

   dprintf(SMERR_DETAIL,"closing file....\n");

   sprintf(outbuff, "\n.\n");
   send_remote(s, outbuff);
   if (!check_remote_response(s, "220")) goto end_send_attempt;
        
end_send_attempt:

   unlink(filename);
   return(0);
}
 
/**
*
*   Copied from E-mail.h (can't include it, too much C++ stuff).
*
**/
#define B_MAIL_ATTR_NAME            "MAIL:name"                 /* indexed string*/
#define B_MAIL_ATTR_STATUS          "MAIL:status"               /* indexed string*/
#define B_MAIL_ATTR_PRIORITY        "MAIL:priority"             /* indexed string*/
#define B_MAIL_ATTR_TO              "MAIL:to"                   /* indexed string*/
#define B_MAIL_ATTR_CC              "MAIL:cc"                   /* indexed string*/
#define B_MAIL_ATTR_FROM            "MAIL:from"                 /* indexed string*/
#define B_MAIL_ATTR_SUBJECT         "MAIL:subject"              /* indexed string*/
#define B_MAIL_ATTR_REPLY           "MAIL:reply"                /* indexed string*/
#define B_MAIL_ATTR_WHEN            "MAIL:when"                 /* indexed time*/
#define B_MAIL_ATTR_FLAGS           "MAIL:flags"                /* indexed int32*/
#define B_MAIL_ATTR_RECIPIENTS      "MAIL:recipients"           /* string*/
#define B_MAIL_ATTR_MIME            "MAIL:mime"	                /* string*/
#define B_MAIL_ATTR_HEADER          "MAIL:header_length"        /* int32*/
#define B_MAIL_ATTR_CONTENT         "MAIL:content_length"       /* int32*/

#define B_MAIL_TO                   "To: "
#define B_MAIL_CC                   "Cc: "
#define B_MAIL_BCC                  "Bcc: "
#define B_MAIL_FROM                 "From: "
#define B_MAIL_DATE                 "Date: "
#define B_MAIL_REPLY                "Reply-To: "
#define B_MAIL_SUBJECT              "Subject: "
#define B_MAIL_PRIORITY             "Priority: "
#define B_MAIL_MIME                 "Mime-Version: "

char *concat(char *s, char *t)
{
   int i;
   char *r;
/*
   if (s) printf("s: %s\n", s);
   if (t) printf("t: %s\n", t);
*/
   r = (char *) malloc((s ? strlen(s) : 0) + (t ? strlen(t) : 0) + 2);
   if (r) {
      r[0] = '\0';
      if (s) {
/*
*   Strip cr/lf's
*/
         i = strlen(s);
         while ((i > 0) && ((s[i - 1] == '\n') || (s[i - 1] == '\r'))) {
            s[i - 1] = '\0';
            i--;
         } 
         strcpy(r, s);
         free(s);
      }
      if (t) strcat(r, t);
   }
/*
   if (r) printf("r: %s\n", r);
*/
   return(r);
}

/**
*
*   Parse the file and set attributes
*
**/
void SetParsedAttrs(FILE *fp)
{
   char *b, *s, *ptr, *ptr1, *ptr2, *ptr3, *ptr4;
   int32 len, l, i, j, blank;
   int32 header_size, header_lines, body_size;
   time_t t;

   static struct {
      char *parse;
      char *name;
      long  id;
   } itemarray[] = {
      {B_MAIL_PRIORITY, B_MAIL_ATTR_PRIORITY, 0 },
      {B_MAIL_TO,       B_MAIL_ATTR_TO,       1 },
      {B_MAIL_CC,       B_MAIL_ATTR_CC,       2 },
      {B_MAIL_FROM,     B_MAIL_ATTR_FROM,     3 },
      {B_MAIL_SUBJECT,  B_MAIL_ATTR_SUBJECT,  4 },
      {B_MAIL_REPLY,    B_MAIL_ATTR_REPLY,    5 },
      {B_MAIL_MIME,     B_MAIL_ATTR_MIME,     6 },
      {B_MAIL_DATE,     B_MAIL_ATTR_WHEN,     7 },
      {NULL,            NULL,                 8 },
   };

   int fd = fileno(fp);
/*
*   Apparently opening the attr dir is required
*/
   DIR *d = fs_fopen_attr_dir(fd);
/*
*   Status = New
*/
   fs_write_attr(fd, B_MAIL_ATTR_STATUS, B_STRING_TYPE, 0, "New", 4);
/*
*   Flags = 0
*/
   i = 0;
   fs_write_attr(fd, B_MAIL_ATTR_FLAGS, B_INT32_TYPE, 0, &i, 4);
/*
*   Find the end of the header.
*/
   rewind(fp);

   header_lines = 0;
   header_size  = 0;

#define BSIZE 16384
   b = (char *) malloc(BSIZE);
   while (fgets(b, BSIZE, fp)) {
/*
*   See if this is a blank line
*/
      blank = 1;
      for (i = 0; i < strlen(b); i++) {
         if ((b[i] != ' ') && (b[i] != '\t') && (b[i] != '\r') && (b[i] != '\n')) blank = 0;
      }
      if (blank) break;

      header_lines++;
   }
/*
*   Header size
*/
   header_size = ftell(fp);
   fs_write_attr(fd, B_MAIL_ATTR_HEADER, B_INT32_TYPE, 0, &header_size, 4);
/*
*   Parse the file for the various header items.
*/
   rewind(fp);

   s = NULL;
   fgets(b, BSIZE, fp);
   for (i = 0; i < header_lines; i++) {
      if (s) free(s);
      s = NULL;
      while (1) {
         s = concat(s, b);
         fgets(b, BSIZE, fp);
         if ((i >= header_lines) || ((*b != ' ') && (*b != '\t'))) break;
         i++;
      }

      for (j = 0; itemarray[j].name; j++) {
         l = strlen(itemarray[j].parse);
         if (strncasecmp(s, itemarray[j].parse, l) == 0) {
/*
*   Special handling of case 3, From: field.
*/
            if (itemarray[j].id == 3) {
               ptr = s + l;
               while ((*ptr == ' ') || (*ptr == '\t')) ptr++;
               ptr1 = ptr;
               ptr3 = ptr;
               while ((*ptr3 != '<') && (*ptr3 != '\0') && (*ptr3 != '\r') && (*ptr3 != '\n')) ptr3++;
               ptr2 = ptr3;
               if ((*ptr3 == '\0') || (*ptr3 == '\r') || (*ptr3 == '\n')) {
                  ptr3 = ptr1;
                  ptr4 = ptr2;
               }
               else {
                  ptr2--;
                  ptr3++;
                  ptr4 = ptr3;
                  while ((*ptr4 != '>') && (*ptr4 != '\0') && (*ptr4 != '\r') && (*ptr4 != '\n')) ptr4++;
               }
               *ptr4 = '\0';
               *ptr2 = '\0';
               fs_write_attr(fd, itemarray[j].name, B_STRING_TYPE, 0, ptr3, ptr4 - ptr3 + 1);
               fs_write_attr(fd, B_MAIL_ATTR_NAME, B_STRING_TYPE, 0, ptr1, ptr2 - ptr1 + 1);
            }
/*
*   Special handling of case 7, Date: field
*/
            else if (itemarray[j].id == 7) {
               ptr = s + l;
               l = strlen(ptr) - 2;
               ptr[l] = '\0';
               t = parsedate(ptr, -1);
               if (t == -1) t = real_time_clock();
/*
*   Note: a B_TIME_TYPE should really be a bigtime_t, but in practice for
*   e-mail it turns out to be just a time_t.
*/
               fs_write_attr(fd, B_MAIL_ATTR_WHEN, B_TIME_TYPE, 0, &t, 4);
            }
            else {
               ptr = s + l;
               l = strlen(ptr) - 2;
               ptr[l] = '\0';
               fs_write_attr(fd, itemarray[j].name, B_STRING_TYPE, 0, ptr, l + 1);
            }
         }
      }
   }
/*
*   Find body size
*/
   fseek(fp, 0, SEEK_END);
   body_size = ftell(fp) - header_size;
   fs_write_attr(fd, B_MAIL_ATTR_CONTENT, B_INT32_TYPE, 0, &body_size, 4);

   if (d) fs_close_attr_dir(d);
   return;
}

/**
*
*   Function to print formatted debug output to stdout.
*
**/
void dprintf(long severity, char *fmt, ...)
{
   va_list args;
   char buf[256];
   unsigned long number;
   struct tm *tm;
   char datestamp[40];

   if (severity < infolevel) {
      va_start(args, fmt);
      vsprintf(buf, fmt, args);
      va_end(args);

      if (smtplog != NULL) {
         number = time(NULL);
         tm = localtime(&number);
         strftime(datestamp , sizeof(datestamp), "%a, %d %b %Y %H:%M:%S", tm);
         fprintf(smtplog, "%s: %s", datestamp, buf);
         fflush(smtplog);
      }
      else {
         printf(buf);
      }
   }
   return;
}

/**
*
*   Function to print long lines broken beyond 80 at a space
*   boundary.
*
**/
void fwritebr(const char *s, FILE *fp, int breakat)
{
   long l;
   const char *bptr = s;
   
   while (1) {
      l = strlen(bptr);
      if (l == 0) break;
/*
*   Short string: just print it
*/
      if (l <= breakat) {
         if (bptr != s) fwrite(" ", 1, 1, fp);
         fwrite(bptr, 1, l, fp);
         fwrite("\r\n", 1, 2, fp);
         break;
      }
/*
*   String longer than breakat characters. Go to character
*   breakat + 1(!) and find the first space at or before that.
*   If there are no spaces, then we break after breakat (we
*   know that breakat + 1 is not a space, and the first character
*   of this string also cannot be a space). We skip multiple
*   spaces afterwards.
*/
      for (l = breakat; l >= 0; l--) {
         if (bptr[l] == ' ' || l == 0) {
            if (l == 0) l = breakat;

            if (bptr != s) fwrite(" ", 1, 1, fp);
            fwrite(bptr, 1, l, fp);
            fwrite("\r\n", 1, 2, fp);

            bptr = &bptr[l];
            while (*bptr == ' ') bptr++;

            break;
         }
      }
   }
   return;
}

int32 process_connection(void *data)
{
   thread_data *my_data = (thread_data*)data; 
   int s = my_data->sock;
   struct sockaddr_in client = my_data->sa;
   int n, sel_result=0, remote = 0;
   char *client_ip = inet_ntoa(client.sin_addr);
   char bf[MAXLINE + 16], *buff, *ptr1;
   struct passwd  *pw;
   char user[MAXHOSTNAMELEN + 4];
   char username[MAXHOSTNAMELEN + 4];
   char hostname[MAXHOSTNAMELEN + 4];
   char fpath[MAXPATHLEN + 4];
   char rpath[MAXPATHLEN + 4];
   long len, mailnum, started, idx, ll;
   struct sockaddr_in remote_server;
       
   fd_set read;

   buff = &bf[5];

   user[0]     = 0;
   username[0] = 0;
   hostname[0] = 0;
   fpath[0]    = 0;
   rpath[0]    = 0;
   bf[0]       = 0;
   buff[0]     = 0;

/*
*   Start SMTP protocol
*/
   putline(s, "220 %s - BeOS System %s Ready", myhostname, &version_txt[6]);
   while (1) {
      n = getline(s, bf);
      if (n <= 0) goto cleanup;

/*
*   Find which command is in the first four characters
*/
      if (n < 4) {
         idx = -1;
      }
      else {
         bf[4] = '\0';
         for (ll = 0; ll < 4; ll++) bf[ll] = toupper(bf[ll]);
         idx = *(unsigned long *) bf;
         idx = ntohl(idx); 
         len = strlen(buff);
      }
/*
*   Handle each command
*/
      switch (idx) {
         case HELO :
            user[0]     = 0;
            username[0] = 0;
            hostname[0] = 0;
            fpath[0]    = 0;
            rpath[0]    = 0;

            if (len == 0) {
               putline(s, "501 No host name");
            }
            else if (len > 64) {
               putline(s, "500 Line too long");
            }
            else {
               struct hostent *he;
               ulong addr = client.sin_addr.s_addr;
               int ok = nochecking ? 1 : 0;
               strcpy(hostname, buff);
               if (ptr1 = strchr(hostname, '\r')) *ptr1 = 0;
               if (ptr1 = strchr(hostname, '\n')) *ptr1 = 0;
/*
*   accept any connections from this machine
*/
               if (strcmp(client_ip, "127.0.0.1")== 0) ok = 1;
/*
*   try to stop localhost connections from remote machines
*/
               if (strcasecmp(hostname, "localhost") == 0) {
                  if (strcmp(client_ip,"127.0.0.1") != 0) ok = 0;
               }

               he = gethostbyaddr((const char*) &addr, 4, AF_INET);
               if (he != NULL && ok == 0) {
                  if (strcasecmp(hostname, he->h_name) == 0) ok = 1;
                  else                                       ok = 2;
               }

               if (ok == 1) {
                  started = 1;
                  putline(s, "250 %s - Hello %s [%s] pleased to meet you!", myhostname, hostname, client_ip);
               }
               else if (ok == 2) {
                  putline(s, "554 - host [%s] doesn't resolve to %s!", client_ip, hostname);
               }
               else {
                  putline(s, "554 - couldn't find details for [%s].  IP address must resolve!", client_ip);
               }
            }
            break;
         case MAIL :
            if (extractpath(rpath, buff)) started = 2;
            putline(s, buff);
            break;
         case RCPT :
            if (started < 2) {
               putline(s, "503 No valid MAIL command received");
            }
            else if (started > 2) {
               putline(s, "452 Single-user system, multiple recipients illogical");
            }
            else {
               if (extractpath(fpath, buff)) {
                  strcpy(username, fpath);
                  strcpy(user, fpath);
                  ptr1 = strchr(user, '@');
                  if (ptr1 == NULL) {
                     strcpy(buff, "553 Incorrect mail box syntax");
                  }
                  else {
                     *ptr1++ = 0;
                     if (strcasecmp(ptr1, myhostname) == 0) {
                        if (nousercheck ||
                            ((pw = getpwnam(user)) && (strcasecmp(pw->pw_name, user) == 0))) {
                           strcpy(buff, "250 OK");
                           started = 3;remote = 0;
                        }
                        else {
                           strcpy(buff, "550 No such user");
                        }
                     }
                     else {
                        struct hostent *he = get_mail_host(ptr1);
                        int port = 25;
                        if (he != NULL) {
                           memcpy((char*) &remote_server.sin_addr, he->h_addr , he->h_length);
                           remote_server.sin_family = AF_INET;
                           remote_server.sin_port = htons(port);
                           memset(remote_server.sin_zero, 0, sizeof(remote_server.sin_zero));
                           remote = 1;
                           started = 3;
                           strcpy(buff, "251 Relaying OK - message being forwarded to ");
                           strcat(buff, he->h_name);
                        }
                        else {
                           strcpy(buff,"551 User not local");
                        }
                     }
                  }
               }
               putline(s, buff);
            }
            break;
         case DATA :
            if (started != 3) {
               putline(s, "503 No recipient");
            }
            else {
               char *tfile;
               putline(s, "354 OK");
               if ((tfile = collectfile(s, maildir, myhostname, rpath, hostname, username, buff)) != NULL) {
                  putline(s, "250 OK");
                  if (remote) {
                     putline(s, "sending to remote host");
                     if (do_remote_mail(tfile, rpath, username, &remote_server) < 0) {
                        putline(s, "remote mail sending failed!");
                     }
                  } 
                  else {
                     do_mail_local(tfile);
                     mailbeep(soundfile);
                  }
               }
               else {
                  putline(s, "554 Error collecting data");
               }
               started = 2;
            }
            break;
         case RSET :
            user[0]     = 0;
            username[0] = 0;
            hostname[0] = 0;
            fpath[0]    = 0;
            rpath[0]    = 0;
            bf[0]       = 0;
            buff[0]     = 0;
            started     = 0;
/* Fall through */
         case NOOP :
            putline(s, "250 OK");
            break;
         case DBUG :
            infolevel = atol(buff);

            if (infolevel < 0) infolevel = 0;
            if (infolevel > 7) infolevel = 7;
/* Fall through */
         case SEND :
         case SOML :
         case SAML :
         case VRFY :
         case EXPN :
         case TURN :
            putline(s, "502 Command not implemented");
            break;
         case HELP :
            putline(s, "211-Available Commands:");
            putline(s, "211 DATA  HELO  HELP  MAIL  NOOP  QUIT  RCPT  RSET");
            break;
         case QUIT :
            putline(s, "221 %s Service closing transmission channel", myhostname);
            goto cleanup;
            break;
         default   :
            putline(s, "502 Command not recognized");
            break;
      }
   }

cleanup:
   closesocket(s);
   sleep(1);
   return;
}


